@@ -26,6 +26,7 @@
 #include "itemlibrarywidget.h"
 
 #include "customfilesystemmodel.h"
+#include "itemlibraryassetimportdialog.h"
 
 #include <theme.h>
 
@@ -36,6 +37,8 @@
 #include <model.h>
 #include <rewritingexception.h>
 #include <qmldesignerplugin.h>
+#include <qmldesignerconstants.h>
+#include <designeractionmanager.h>
 
 #include <utils/algorithm.h>
 #include <utils/flowlayout.h>
@@ -48,6 +51,10 @@
 #include <coreplugin/icore.h>
 #include <coreplugin/messagebox.h>
 
+#ifdef IMPORT_QUICK3D_ASSETS
+#include <QtQuick3DAssetImport/private/qssgassetimportmanager_p.h>
+#endif
+
 #include <QApplication>
 #include <QDrag>
 #include <QFileDialog>
@@ -111,7 +118,7 @@ ItemLibraryWidget::ItemLibraryWidget(QWidget *parent) :
     Theme::setupTheme(m_itemViewQuickWidget->engine());
 
     /* other widgets */
-    QTabBar *tabBar = new QTabBar(this);
+    auto tabBar = new QTabBar(this);
     tabBar->addTab(tr("QML Types", "Title of library QML types view"));
     tabBar->addTab(tr("Resources", "Title of library resources view"));
     tabBar->addTab(tr("Imports", "Title of library imports view"));
@@ -128,8 +135,8 @@ ItemLibraryWidget::ItemLibraryWidget(QWidget *parent) :
     m_filterLineEdit->setFiltering(true);
     QWidget *lineEditFrame = new QWidget(this);
     lineEditFrame->setObjectName(QStringLiteral("itemLibrarySearchInputFrame"));
-    QGridLayout *lineEditLayout = new QGridLayout(lineEditFrame);
-    lineEditLayout->setMargin(2);
+    auto lineEditLayout = new QGridLayout(lineEditFrame);
+    lineEditLayout->setContentsMargins(2, 2, 2, 2);
     lineEditLayout->setSpacing(0);
     lineEditLayout->addItem(new QSpacerItem(5, 3, QSizePolicy::Fixed, QSizePolicy::Fixed), 0, 0, 1, 3);
     lineEditLayout->addItem(new QSpacerItem(5, 5, QSizePolicy::Fixed, QSizePolicy::Fixed), 1, 0);
@@ -145,7 +152,7 @@ ItemLibraryWidget::ItemLibraryWidget(QWidget *parent) :
     spacer->setObjectName(QStringLiteral("itemLibrarySearchInputSpacer"));
     spacer->setFixedHeight(4);
 
-    QGridLayout *layout = new QGridLayout(this);
+    auto layout = new QGridLayout(this);
     layout->setContentsMargins(0, 0, 0, 0);
     layout->setSpacing(0);
     layout->addWidget(tabBar, 0, 0, 1, 1);
@@ -166,12 +173,12 @@ ItemLibraryWidget::ItemLibraryWidget(QWidget *parent) :
 
     connect(&m_compressionTimer, &QTimer::timeout, this, &ItemLibraryWidget::updateModel);
 
-    auto *flowLayout = new Utils::FlowLayout(m_importTagsWidget.data());
-    flowLayout->setMargin(4);
+    auto flowLayout = new Utils::FlowLayout(m_importTagsWidget.data());
+    flowLayout->setContentsMargins(4, 4, 4, 4);
 
     m_addResourcesWidget->setVisible(false);
     flowLayout = new Utils::FlowLayout(m_addResourcesWidget.data());
-    flowLayout->setMargin(4);
+    flowLayout->setContentsMargins(4, 4, 4, 4);
     auto button = new QToolButton(m_addResourcesWidget.data());
     auto font = button->font();
     font.setPixelSize(Theme::instance()->smallFontPixelSize());
@@ -183,6 +190,49 @@ ItemLibraryWidget::ItemLibraryWidget(QWidget *parent) :
     flowLayout->addWidget(button);
     connect(button, &QToolButton::clicked, this, &ItemLibraryWidget::addResources);
 
+#ifdef IMPORT_QUICK3D_ASSETS
+    DesignerActionManager *actionManager =
+             &QmlDesignerPlugin::instance()->viewManager().designerActionManager();
+
+    auto handle3DModel = [](const QStringList &fileNames, const QString &defaultDir) -> bool {
+        auto importDlg = new ItemLibraryAssetImportDialog(fileNames, defaultDir);
+        importDlg->show();
+        return true;
+    };
+
+    auto add3DHandler = [&](const QString &category, const QString &ext) {
+        const QString filter = QStringLiteral("*.%1").arg(ext);
+        actionManager->registerAddResourceHandler(
+                    AddResourceHandler(category, filter, handle3DModel, 10));
+    };
+
+    QSSGAssetImportManager importManager;
+    QHash<QString, QStringList> supportedExtensions = importManager.getSupportedExtensions();
+
+    // All things importable by QSSGAssetImportManager are considered to be in the same category
+    // so we don't get multiple separate import dialogs when different file types are imported.
+    const QString category = tr("3D Assets");
+
+    // Skip if 3D asset handlers have already been added
+    const QList<AddResourceHandler> handlers = actionManager->addResourceHandler();
+    bool categoryAlreadyAdded = false;
+    for (const auto &handler : handlers) {
+        if (handler.category == category) {
+            categoryAlreadyAdded = true;
+            break;
+        }
+    }
+
+    if (!categoryAlreadyAdded) {
+        const auto groups = supportedExtensions.keys();
+        for (const auto &group : groups) {
+            const auto extensions = supportedExtensions[group];
+            for (const auto &ext : extensions)
+                add3DHandler(category, ext);
+        }
+    }
+#endif
+
     // init the first load of the QML UI elements
     reloadQmlSource();
 }
@@ -192,13 +242,19 @@ void ItemLibraryWidget::setItemLibraryInfo(ItemLibraryInfo *itemLibraryInfo)
     if (m_itemLibraryInfo.data() == itemLibraryInfo)
         return;
 
-    if (m_itemLibraryInfo)
+    if (m_itemLibraryInfo) {
         disconnect(m_itemLibraryInfo.data(), &ItemLibraryInfo::entriesChanged,
                    this, &ItemLibraryWidget::delayedUpdateModel);
+        disconnect(m_itemLibraryInfo.data(), &ItemLibraryInfo::importTagsChanged,
+                   this, &ItemLibraryWidget::delayedUpdateModel);
+    }
     m_itemLibraryInfo = itemLibraryInfo;
-    if (itemLibraryInfo)
+    if (itemLibraryInfo) {
         connect(m_itemLibraryInfo.data(), &ItemLibraryInfo::entriesChanged,
                 this, &ItemLibraryWidget::delayedUpdateModel);
+        connect(m_itemLibraryInfo.data(), &ItemLibraryInfo::importTagsChanged,
+                this, &ItemLibraryWidget::delayedUpdateModel);
+    }
     delayedUpdateModel();
 }
 
@@ -291,7 +347,7 @@ void ItemLibraryWidget::setupImportTagWidget()
 
     qDeleteAll(m_importTagsWidget->findChildren<QWidget*>("", Qt::FindDirectChildrenOnly));
 
-    auto *flowLayout = m_importTagsWidget->layout();
+    auto flowLayout = m_importTagsWidget->layout();
 
     auto createButton = [this](const QString &import) {
         auto button = new QToolButton(m_importTagsWidget.data());
@@ -318,6 +374,8 @@ void ItemLibraryWidget::setupImportTagWidget()
 
 void ItemLibraryWidget::updateModel()
 {
+    QTC_ASSERT(m_itemLibraryModel, return);
+
     m_itemLibraryModel->update(m_itemLibraryInfo.data(), m_model.data());
     updateImports();
     updateSearch();
@@ -342,7 +400,7 @@ void ItemLibraryWidget::startDragAndDrop(QQuickItem *mouseArea, QVariant itemLib
     m_currentitemLibraryEntry = itemLibraryId.value<ItemLibraryEntry>();
 
     QMimeData *mimeData = m_itemLibraryModel->getMimeData(m_currentitemLibraryEntry);
-    QDrag *drag = new QDrag(this);
+    auto drag = new QDrag(this);
 
     drag->setPixmap(Utils::StyleHelper::dpiSpecificImageFile(
                         m_currentitemLibraryEntry.libraryEntryIconPath()));
@@ -381,7 +439,22 @@ void ItemLibraryWidget::addPossibleImport(const QString &name)
     QTC_ASSERT(m_model, return);
     const Import import = m_model->highestPossibleImport(name);
     try {
-        m_model->changeImports({Import::createLibraryImport(name, import.version())}, {});
+        QList<Import> addedImports = {Import::createLibraryImport(name, import.version())};
+        // Special case for adding an import for 3D asset - also add QtQuick3D import
+        const QString asset3DPrefix = QLatin1String(Constants::QUICK_3D_ASSETS_FOLDER + 1)
+                + QLatin1Char('.');
+        if (name.startsWith(asset3DPrefix)) {
+            const QString q3Dlib = QLatin1String(Constants::QT_QUICK_3D_MODULE_NAME);
+            Import q3DImport = m_model->highestPossibleImport(q3Dlib);
+            if (q3DImport.url() == q3Dlib)
+                addedImports.prepend(Import::createLibraryImport(q3Dlib, q3DImport.version()));
+        }
+        RewriterTransaction transaction
+                = m_model->rewriterView()->beginRewriterTransaction(
+                    QByteArrayLiteral("ItemLibraryWidget::addPossibleImport"));
+
+        m_model->changeImports(addedImports, {});
+        transaction.commit();
     }
     catch (const RewritingException &e) {
         e.showException();
@@ -429,15 +502,22 @@ void ItemLibraryWidget::addResources()
 
     filters.prepend(tr("All Files (%1)").arg(map.values().join(" ")));
 
+    static QString lastDir;
+    const QString currentDir = lastDir.isEmpty() ? document->fileName().parentDir().toString() : lastDir;
+
     const auto fileNames = QFileDialog::getOpenFileNames(this,
                                                    tr("Add Resources"),
-                                                   document->fileName().parentDir().toString(),
+                                                   currentDir,
                                                    filters.join(";;"));
 
+
+    if (!fileNames.isEmpty())
+        lastDir = QFileInfo(fileNames.first()).absolutePath();
+
     QMultiMap<QString, QString> partitionedFileNames;
 
     for (const QString &fileName : fileNames) {
-        const QString suffix = "*." + QFileInfo(fileName).completeSuffix();
+        const QString suffix = "*." + QFileInfo(fileName).suffix().toLower();
         const QString category = reverseMap.value(suffix);
         partitionedFileNames.insert(category, fileName);
     }
